window.by = window.by || {};

(function($, exports) {
	var meetingTypes = {
		coffee: {
			terms: [ "coffee" ],
			categories: [ "cafes" ],
			range: { start: 9, end: 17 }
		},
		beer: {
			terms: [ "beer" ],
			categories: [ "bars" ],
			range: { start: 16, end: 19 }
		},
		breakfast: {
			terms: [ "breakfast" ],
			categories: [ "breakfast_brunch" ],
			range: { start: 7, end: 10 }
		},
		lunch: {
			terms: [ "lunch" ],
			categories: [],
			range: { start: 11, end: 13 }
		},
		dinner: {
			terms: [ "dinner" ],
			categories: [],
			range: { start: 18, end: 21 }
		}
	};

	// The higher the weight, the more likely it'll be pick for meeting in a box
	var meetingDayWeights = [ 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ];

	var MS_IN_DAY = 24 * 60 * 60 * 1000;
	var BIAS_AGAINST_WEEKENDS = -2;
	function eventToRange(event, day, dayCount) {
		var start = event.start, end = event.end;
		var startHour = start.getHours() + start.getMinutes() / 60 + start.getSeconds() / 3600;
		var endHour = end.getHours() + end.getMinutes() / 60 + end.getSeconds() / 3600;

		if(dayCount < 1) { // Not a multi-day event
			return { start: startHour, end: endHour, isTentative: event.allDay };
		}

		if(day === 0) { // Day of multi-day event
			return { start: startHour, end: 24, isTentative: event.allDay };
		}

		if(day === dayCount - 1) { // Last day of multi-day event
			return { start: 0, end: endHour, isTentative: event.allDay };
		}

		// In the middle of a multi-day event
		return { start: 0, end: 24, isTentative: true };
	}
	function doRangesOverlap(rangeA, rangeB) {
		if(rangeA.start > rangeA.end) { throw new Error("Invalid rangeA. Start greater than end"); }
		if(rangeB.start > rangeB.end) { throw new Error("Invalid rangeB. Start greater than end"); }

		if(rangeA.end < rangeB.start) { return false; }
		if(rangeB.end < rangeA.start) { return false; }

		return true;
	}
	function weightToDay(weight) {
		var now = $.now(), acc = 0;

		for(var i = 0; i < meetingDayWeights.length; i++) {
			var weekendBias = 0, day = new Date(now + MS_IN_DAY * i);
			if(day.getDay() === 0 || day.getDay() === 6) { // is Weekend?
				weekendBias = BIAS_AGAINST_WEEKENDS;
			}

			acc += meetingDayWeights[i] + weekendBias;

			if(weight < acc) { return new Date(day.getFullYear(), day.getMonth(), day.getDate()); }
		}

		throw "A weight too high was passed in -- weight: " + weight;
	}
	function subtractRange(rangeA, rangeB) {
		// rangeA represented by '-'
		// rangeB represented by '|'
		// Where rangeA and rangeB overlap is represented by '+'
		if(!doRangesOverlap(rangeA, rangeB)) { return [ rangeA ]; } // -- ||

		if(rangeB.start <= rangeA.start && rangeA.end <= rangeB.end) { // |++| or ++++ or ++|| or ||++
			return [];
		} else if(rangeA.start < rangeB.start && rangeB.end < rangeA.end) { // -++-
			return [
				{ start: rangeA.start, end: rangeB.start, isTentative: rangeA.isTentative },
				{ start: rangeB.end, end: rangeA.end, isTentative: rangeA.isTentative }
			];
		} else if(rangeA.start < rangeB.start && rangeB.start < rangeA.end) { // --++, --+|
			return [ { start: rangeA.start, end: rangeB.start, isTentative: rangeA.isTentative } ];
		} else { // ++--, |+--
			return [ { start: rangeB.end, end: rangeA.end, isTentative: rangeA.isTentative } ];
		}
	}
	function subtractRanges(rangesA, rangesB) {
		if(rangesA.length === 0) { return []; }
		if(rangesB.length === 0) { return rangesA; }

		var results = rangesA;
		for(var b = 0; b < rangesB.length; b++) {
			var rangeB = rangesB[b];

			var tmp = [];
			results.forEach(function(r) {
				tmp = tmp.concat(subtractRange(r, rangeB));
			});
			results = tmp;
		}

		return results;
	}
	// TODO: bias against allday, weekend; bias to nearer in time, to business hours
	function pickAvailableTimes(options, scheduledEvents) {
		var availableRange = { start: 24, end: 0 };
		
		Object.keys(meetingTypes).forEach(function(type) {
			if(!options[type]) { return; }

			var values = meetingTypes[type];
			if(values.range.start < availableRange.start) { availableRange.start = values.range.start; }
			if(values.range.end > availableRange.end) { availableRange.end = values.range.end; }
		});

		var rangesByDays = {};

		(scheduledEvents || []).forEach(function(event) { // Turn events into ranges and split them by days
			var days = Math.min((event.end - event.start) / MS_IN_DAY, 30);

			for(var day = 0; day < days; day++) {
				var next = new Date(event.start + MS_IN_DAY * day);
				var idx = next.getMonth() + "-" + next.getDate();

				var ranges = rangesByDays[idx] || [];
				var range = eventToRange(event, day, days);

				var didRangesMerged = false;
				for(var i = 0; i < ranges.length; i++) { // Merge ranges if necessary
					if(range.isTentative !== ranges[i].isTentative) { continue; }

					if(!doRangesOverlap(range, ranges[i])) { continue; }

					ranges[i].start = Math.min(ranges[i].start, range.start);
					ranges[i].end = Math.max(ranges[i].end, range.end);
					didRangesMerged = true;
					break;
				}

				if(!didRangesMerged) { ranges.push(range); }

				rangesByDays[idx] = ranges;
			}
		});

		var totalWeight = meetingDayWeights.reduce(function(a, b) { return a + b; }) + BIAS_AGAINST_WEEKENDS * 4;
		var dayPicked, dayIdx, daysConsidered = {}, availableRanges, tries = 0;

		var times = [];
		for(var i = 0; i < (options.timesNeeded || 1); i++) {
			do {
				dayPicked = weightToDay((Math.random() * totalWeight) >> 0);
				dayIdx = dayPicked.getMonth() + "-" + dayPicked.getDate();

				var allRanges = rangesByDays[dayIdx] || [];
				var takenRanges = allRanges.filter(function(r) { return !r.isTentative; });
				var tentativeRanges = allRanges.filter(function(r) { return r.isTentative; });

				daysConsidered[dayIdx] = (daysConsidered[dayIdx] || 0) + 1;

				var curRanges = subtractRanges([ availableRange ], takenRanges);
				availableRanges = curRanges.filter(function(r) { return 1 <= r.end - r.start; });
			} while(availableRanges.length <= 0 && tries++ < 100);

			if(availableRanges.length <= 0) { return null; }

			var rndRange = availableRanges[(availableRanges.length * Math.random()) >> 0];
			var rndHour = (rndRange.start + (rndRange.end - rndRange.start) * Math.random()) >> 0;

			rangesByDays[dayIdx] = rangesByDays[dayIdx] || [];
			rangesByDays[dayIdx].push({ start: rndHour, end: rndHour + 1 }); // Prevents duplicate picks

			times.push({
				start: new Date(dayPicked.setHours(rndHour)),
				end: new Date(dayPicked.setHours(rndHour + 1))
			});
		}
		return times;
	}

	function findCityCoords(city, callback) {
		var url = "https://boomerangcalendar.baydin.com/findcitycoordinates?city=" + city;
		function what_to_do_after_making_google_api_call(data){
			data = JSON.parse(data);
			if(data.status !== "OK") {
				return callback({ type: "geocode-failed-city", status: data.status, city: city });
			}

			if(data.results.length <= 0) {
				return callback({ type: "geocode-no-results-city", results: data.results, city: city });
			}

			var location = data.results[0].geometry;
			callback(null, { lat: location.location.lat, lng: location.location.lng });
		}
		send_get_request(url, what_to_do_after_making_google_api_call, false, false);

	}

	var FETCH_VENUE_COUNTER = 0; // TODO: refactor this whole flow. this is a hack to avoid infinite geocoding loop

	function fetch_venue(options, scheduledEvents, callback) {
		if(!options.city && (options.lat == null || options.lng == null)) {
			return setTimeout(function() { callback({ type: "no-city-or-coords" }); }, 0);
		}

		if(options.city && (options.lat == null || options.lng == null)) {
			return findCityCoords(options.city, function(err, coords) {
				if(err) { callback(err); }

				options.lat = coords.lat;
				options.lng = coords.lng;

				FETCH_VENUE_COUNTER += 1;
				if (FETCH_VENUE_COUNTER < 5){
					fetch_venue(options, scheduledEvents, callback);
				}
				else{
					alert("We couldn't find a venue for your meeting. You will have to figure it out on your own."); 
					FETCH_VENUE_COUNTER = 0;
				}
			});
		}

		var terms = [], categories = [];
		
		Object.keys(meetingTypes).forEach(function(type) {
			if(!options[type]) { return; }

			var values = meetingTypes[type];
			terms = terms.concat(values.terms);
			categories = categories.concat(values.categories);
		});

		FETCH_VENUE_COUNTER = 0;

		var url = "https://boomerangcalendar.baydin.com/yelpsearch?" +
			"guser=" + encodeURIComponent(get_gmail_user()) +
			"&terms=" + encodeURIComponent(terms.join(" ")) +
			"&latitude=" + encodeURIComponent(options.lat) +
			"&longitude=" + encodeURIComponent(options.lng) +
			"&categories=" + encodeURIComponent(categories.join())
		;

		send_get_request_auth_wrapper(url, function(data){
			data = JSON.parse(data);
			if(!data["businesses"] || data["businesses"].length === 0) {
				return callback({ type: "yelp-no-businesses-found" });
			}

			var idx = (Math.random() * data["businesses"].length) >> 0;
			callback(null, data["businesses"][idx]);
		});
	}

	function get_city_from_latlng(lat, lng, callback) {
		var url = "https://boomerangcalendar.baydin.com/geocode?latlng=" + lat + "," + lng;
		function what_to_do_after_making_google_api_call(data){
			data = JSON.parse(data);
			if(data.status !== "OK") {
				return callback({ type: "geocode-failed-coords", status: data.status, lat: lat, lng: lng });
			}

			if(data.results.length <= 0) {
				return callback({ type: "geocode-no-results-coords", results: data.results, lat: lat, lng: lng });
			}

			var city;
			for (var i = 0; i < data.results[0].address_components.length; i++) {
				var address = data.results[0].address_components[i];

				for (var b = 0; b < address.types.length; b++) {
					//there are different types that might hold a city admin_area_lvl_1
					// usually does in come cases looking for sublocality type will be more appropriate
					if (address.types[b] === "locality") {
						//this is the object you are looking for
						city = address;
						break;
					}
				}
			} 
			if(city) { callback(null, city); }
			else { callback({ type: "geocode-no-locality-found" }); }
		}
		send_get_request(url, what_to_do_after_making_google_api_call, false, false);

	}

	function get_city_from_browser(callback) {
		if(!navigator.geolocation) {
			// Make error returning aync as well
			return setTimeout(function() { callback({ type: "geolocation-not-supported" }); }, 0);
		}

		navigator.geolocation.getCurrentPosition(
			function showLocation(position) {
				var latitude = position.coords.latitude;
				var longitude = position.coords.longitude;

				get_city_from_latlng(latitude, longitude, function(err, city) {
					if(err)	{ throw err; }

					callback(null, { city: city.long_name, lat: latitude, lng: longitude });
				});
			},
			function errorHandler(err) {
				if(err.code === 1) { return callback({ type: "geolocation-access-denied" }); }
				if(err.code === 2) { return callback({ type: "geolocation-position-unavailable" }); }

				return callback({ type: "geolocation-unknown-" + err.code, err: err });
			},
			{ timeout: 3000 } // 3 seconds
		);
	}

	exports.pickAvailableTimes = pickAvailableTimes; // TODO: write unit tests
	exports.fetch_venue = fetch_venue;
	exports.get_city_from_browser = get_city_from_browser;
})(jQuery, window.by.randomizer || (window.by.randomizer = {}));
